% File <determine_Mfield_erlang_bounds.m>
%
% Author: Falko Bause
% Date: June 2016   changed February 2018 (storage efficient representation of Ak)
%
% This file is part of the MATLAB functions for fitting mixture distributions
% Copyright (C) Falko Bause, Department of Computer Science, TU Dortmund, 2016.
%


function [ M,Ak,inivec] = determine_Mfield_erlang_bounds(tracemoments, klist, maxmoments, values, maxelem, M, Ak, inivec)

% Function to determine a field of Erlang distributions.
% M contains the moment vectors of the Erlang dists
% Ak cells containing the generators of the Erlang distributions  (in the form (rate number_of_phases))
% inivec cells containing the initial probability vectors of the Erlang
%           dists which are all (1 number_of_phases); inivec is only here for
%           historical reasons

% maxelem is a number limiting the maximum value of a moment, i.e. if some
% element in the moment vector exceeds maxelem then this moment vector is
% not inserted into moment matrix M and also the erlang dist is not
% inserted into Ak and inivec.

% values is a list of values [a1,a2,a3...] and ai is the mean value of the
% inserted erlang dist. Several erlang dists with different number of
% states (with values from list klist) are generated for the same mean value
% and are inserted into the list.
% If klist contains zero also a deterministic distribution with the given mean values
% is inserted.

% If this function is called with initially non-empty lists M,Ak,inivec these entries
% are kept in the lists.


delmvcounter = 0;

if (nargin <= 5)
    Ak = {};
    inivec = {};
    M = [];
    counter = 1;
else
    counter = size(M,2) + 1;
end

for i=1:length(values)
    
    meanval = values(i);
    
    if (meanval > 0)
        for states = klist
            
            if (states > 0)
                lambda = (1.0*states) / meanval;
            else
                lambda = 1.0/meanval;
            end
            
            if (states > 0)
                mv = momvec_erlang(lambda,states,maxmoments);
            else
                mv = momvec_deterministic(lambda,maxmoments);
            end
            normedmv = (mv./tracemoments);        
            if ( (max(normedmv) < 1.0e-08) || (max(normedmv) > 1.0e+08))
                delmvcounter = delmvcounter + 1;
                % fprintf('INFO: min(normedmv) = %E; max(normedmv) = %E   delmvcounter = %d\n',min(normedmv),max(normedmv),delmvcounter);
                continue;    % components moment vector is too far from the trace moments   
            end
            
            if ((max(mv) <= maxelem) && (not_in_Aks(Ak,size(M,2),lambda,states,1.0e-6)))
                M = horzcat(M,mv');
                if (states > 0)
                    [Aktmp,inivectmp] =  Akinivec_for_Erlang_dists(mv',states);
                else
                    [Aktmp,inivectmp] =  Akinivec_for_deterministic_dists(mv');
                end
                Ak{counter} = Aktmp{1};
                inivec{counter} = inivectmp{1};
                counter = counter + 1;
            end
            if (max(mv) > maxelem)
               fprintf('WARNING: Moments vector element (%E) of Erlang distribution is greater than maxelem = %E !!!\n',max(mv),maxelem);
               fprintf('Erlang distribution is NOT inserted!!\n\n');
            end
        end
        
    end
end
fprintf('Function determine_Mfield_erlang_bounds: %d #components have not been inserted, since normed moment vector is out of bounds.\n Total number of relevant components is %d.\n\n',delmvcounter,counter-1);
