% File <epdf_and_cdfatvals.m>
%
% Author: Falko Bause
% Date: June 2016
%
% This file is part of the MATLAB functions for fitting mixture distributions
% Copyright (C) Falko Bause, Department of Computer Science, TU Dortmund, 2016.
%


function [epdfvec, ecdfvec, pdftvalues, xval] = epdf_and_cdfatvals(trace,xval)

% calculates the (approximate) empirical cdf (cumulative density function) at points xval
% using Matlab function ecdf for the given trace
% and calculates the empirical pdf of the trace at points determined by the Freedman/Diaconis rule,
% the corresponding points are stored in pdftvalues

% Note that the representation is a critical aspect, since the fitting result depends on the PDF representation.
% This aspect becomes relevant when comparing the results with other fitting tools, since they might have
% a different PDF representation, which is essentially based on a different number of histograms

    
[f,x] = ecdf(trace);
N = length(xval);
ecdfvec = zeros(1,N);

for i=1:N
    ecdfvec(i) = max(f(x<=xval(i)));
end



% Calculate empirical pdf:

% % delete outliers!?
del_outliers = 1;
Ltrold = length(trace);
if del_outliers
    qtr = quantile(trace,1000);
    maxtmp = qtr(1000);
    mintmp = qtr(1);
    trace = trace(trace < maxtmp & trace > mintmp);
end

mintr = min(trace);
maxtr = max(trace);
Ltr = length(trace);
fprintf('Original trace length = %d, length after deleting outliers = %d  ===>  ratio is %E\n',Ltrold,Ltr,(Ltr/Ltrold));


% Determine number of bins for PDF histogram using Freedman/Diaconis rule:
rangetr = iqr(trace);  % Determine interquartile range of the trace
                       % https://en.wikipedia.org/wiki/Freedman%E2%80%93Diaconis_rule
                       % https://en.wikipedia.org/wiki/Histogram
binsize = 2*rangetr / (Ltr^(1/3));
no_pdfpoints = round((maxtr-mintr)/binsize);
fprintf('no_pdfpoints after applying Freedman-Diaconis rule %d         (binsize = %E) \n',no_pdfpoints,binsize);

no_pdfpoints = max(min(no_pdfpoints,2000),200);  % changing number of points to a reasonable size if necessary, here between 200 and 2000
                                                % depends on trace characteristics and might have to be changed!!
fprintf('setting no_pdfpoints to  %d   !!!!!!!!!!! \n',no_pdfpoints);


% set no_pdfpoints for later while loop:
no_pdfpoints = 2*no_pdfpoints;
no_pdfzeros = no_pdfpoints;

zeropercent = 0.99999; % value should be between ]0,1[ (excluding limit values!)

% Reduce the number of intervals/pdfpoints if more than "zeropercent" percent of the
% pdf values in the midth of the distribution (see below) are zero;
% if this is the case the number of intervals/points is too large or the
% trace is too short. Anyway the epdfvec vector would probably not contain an
% adequate approximation of the distribution.
% Note, that it is usually a problem if the PDF touches the ground and
% we consider it unlikely to happen, so we reduce the number of points
% for pdf fitting. In the worst case this might result in an imprecise
% representation of the empirical distribution.

while ((no_pdfzeros/no_pdfpoints) > zeropercent) 
% while loop is entered at least once!  (provided zeropercent < 1)
% one loop execution is necessary to define and fill vectors pdftvalues and epdfvec, which store the x,y-values of the PDF function
  
    no_pdfpoints = round(no_pdfpoints/2);
    
    epdfvec = zeros(1,no_pdfpoints);
    pdftvalues = zeros(no_pdfpoints,1);
    
    width = (maxtr-mintr)/no_pdfpoints
    
    for j=1:no_pdfpoints
        pdftvalues(j) = mintr + j*width - 1/2*width; % tvalue is the midth of the interval
    end
    for i=1:Ltr
        ind = floor((trace(i) - mintr)/width); % ind ranges from 0 to no_pdfpoints-1
        if ((ind < no_pdfpoints))
            % increase ind in order to increase the correct "bin" in epdfvec,
            % only the value maxtr would result in an index overflow so we
            % don't increase if trace(i)==maxtr
            ind = ind + 1;
        end
        epdfvec(ind) = epdfvec(ind) + 1;
    end
    % Count (approximately) the number of zero elements within the "midth" of the empirical pdf
    no_pdfzeros = round(0.8*length(epdfvec) - nnz(epdfvec(round(0.1*length(epdfvec)):round(0.9*length(epdfvec)))));
end
epdfvec = epdfvec./(Ltr*width);

fprintf('no_pdfpoints after applying %E percent correction:  %d    \n',zeropercent,no_pdfpoints);


% simple moving average smoothing of PDF curve
% might be useful to eliminate unwanted peaks in PDF representation
% movstepsize = 3;
% epdfvec = smooth(epdfvec,movstepsize,'moving');
% epdfvec = epdfvec';



% Check whether area is approx. one:
area = 0.0;

for i=1:no_pdfpoints
    area = area + (epdfvec(i) * width);
end




fprintf('epdf_and_cdfatvals: area = %E  (should be approx. one!!)\n',area);

fprintf('epdf_and_cdfatvals: mean(epdfvec) = %E, median(epdfvec) = %E, number of non-zeros in epdfvec = %d, length(epdfvec) = %d  \n',mean(epdfvec),median(epdfvec),nnz(epdfvec),length(epdfvec));


% prtmat = [pdftvalues epdfvec'];
% fprintf('====== pdftvalues   epdfvec   Start ========= \n');
% disp(prtmat);
% fprintf('====== pdftvalues   epdfvec   End ========= \n');
