% File <fitpdf_and_cdf_withmoments.m>
%
% Author: Falko Bause
% Date: June 2016
%
% This file is part of the MATLAB functions for fitting mixture distributions
% Copyright (C) Falko Bause, Department of Computer Science, TU Dortmund, 2016.
%



function [pivec,M,inivec,Ak,N,norm2_error,exitflag] = fitpdf_and_cdf_withmoments(tracemoments, tracepdf, tracecdf, pdftvalues, cdftvalues, pivec_initialguess, inivecbestall, Akbestall, Mbestall, Kdist, norm2_error_threshold, options,pdfcdffit_only,jvm,fpstates)

% This function fits the mixture probabilites for a given set of
% Erlang distributions using the pdf, cdf and moments of a trace

% The resultant Erlang distributions are stored in Ak,inivec, M.
% The fitted mixture probabilites are in vector pivec.

% The approach uses NNLS and it is assumed that the pdf,cdf values of the
% trace are given for the x-values in vectors pdftvalues and cdftvalues.


P = []; % matrix for later NNLS call.

no_list_elems = size(Mbestall,2);


% Determine matrix Pcdf:
Pcdf = [];

for i=1:no_list_elems
    Pcdf = vertcat(Pcdf,(erlang_dtm_cdf(Akbestall{i},cdftvalues')));
end


% Determine matrix Ppdf:
Ppdf = [];

for i=1:no_list_elems
    Ppdf = vertcat(Ppdf,(erlang_dtm_pdf(Akbestall{i},pdftvalues')));
end

% Disadvantaging Erlang distributions with a large number of states:
% fpstates is a weight factor determining the influence of the penalty in comparison to fitting of PDF/CDF/Moments in the NNLS problem definition.
Pstates = [];
for i=1:no_list_elems
    Pstates = vertcat(Pstates,Akbestall{i}(1,2));
end
Pstates = fpstates * (ones(no_list_elems,1) + Pstates ./ sum(Pstates));  % Note that sum_i pivec_i = 1 or sum_i fpstates*pivec_i = fpstates resp.

% insert the moments of the Erlang dists into P weighted by the inverse of the emprirical
% moments and an addtional weighting factor f2. Thus also the later d vector has to be increased by 1/f2 values.
% The weighting factor should ensure that the large values of the moments
% do not dominate the fitting of the pdf/cdf values.


f2mean = Kdist; % To ensure that the mean value is fitted well
                % if you change this value you have to change also 
                % other values below

% Further weighting factors:

% #Eqns for PDF/CDF = length(tvalues)
% #Eqns for Moments = Kdist

f1 = 1.0/(mean(tracepdf)*(length(pdftvalues)));
f3 = 1.0/(mean(tracecdf)*(length(cdftvalues)));
f2 = 1.0/(Kdist); % Note that the moments are normalized to (1,1,...,1)

f2 = f2 / (((Kdist+1)/2.0) * 1.0); % because of f2mean=Kdist and the following emphasis on the first moments

momfactor = 1+1/log(max(Kdist,2));  % factor stressing moment fitting if the number of moments is small,
                             % factor is approx. 1 for a large number of moments preventing imbalanced NNLs problem definitions; 
                             % weights for PDF/CDF are 1 by default. max(Kdist,2) just to prevent division by zero.
f2 = f2 * momfactor ; 

if (pdfcdffit_only > 0)
% specific values for specific fittings:
      f2 = f2 / (Kdist); % use additionally only this line for a mere fitting of PDF/CDF,  moments are not fitted (well)   
end 
 
fprintf('!!!!!!!!!!!! f1 for PDF = %E,  f2 for moms = %E,  f2mean = %E,  f3 for CDF = %E, fpstates = %E\n',f1,f2,f2mean,f3,fpstates);

Mbesttmp = Mbestall;
Mbesttmp(1,:) = []; % delete the zero-th moments which are all 1
% Mbesttmp = (f2)*(Mbesttmp'*inv(diag(tracemoments(2:(Kdist+1))))); % better is
Mbesttmp = (f2)*(Mbesttmp'*(diag(ones(1,Kdist)./tracemoments(2:(Kdist+1)))));
for momzaehler=1:Kdist
    Mbesttmp(:,momzaehler) = (f2mean+1-momzaehler)/(1.0)*Mbesttmp(:,momzaehler); % just to emphasize the first moments, so that the mean, 2nd moment etc. of the trace should be fitted well.
end

P = horzcat((f1)*Ppdf,(f3)*Pcdf,Mbesttmp);
P = horzcat(P,Pstates);
d = vertcat((f1)*tracepdf,(f3)*tracecdf);

% increase d by ones weighted by the above calculated weighting factors:
momd = (f2)*ones(Kdist,1);
for momzaehler=1:Kdist
    momd(momzaehler) = (f2mean+1-momzaehler)/(1.0)*momd(momzaehler); % results from emphasizing the first moments, see above
end
d = vertcat(d,momd);
d = vertcat(d,fpstates);

fprintf('min(d) = %E,  max(d) = %E, mean(d) = %E, median(d) = %E, sum(d(PDF)) = %E, sum(d(CDF)) = %E, sum(d(MOM)) = %E\n',min(d),max(d),mean(d),median(d),sum(d(1:length(pdftvalues))),sum(d(length(pdftvalues)+1:(length(pdftvalues)+length(cdftvalues)))),sum(d((length(pdftvalues)+length(cdftvalues))+1:(length(d)-1))));

C = P';
Aeq = ones(1,no_list_elems);
beq = [1];

lb = zeros(no_list_elems,1); % lower bound vector
ub = ones(no_list_elems,1);  % upper bound vector
% OLD: pivec_initialguess = (ones(no_list_elems,1) / no_list_elems);

fprintf('Start lsqlin: (solving NNLS problem) ');
printtime();
[pivecres,~,~,exitflag,~,~,norm2_error] = lsqlin_own_version(C,d,[],[],Aeq,beq,lb,ub,pivec_initialguess,options,1);
tmpvec = C*pivecres-d;
norm2_error = norm(tmpvec(1:(length(tmpvec)-1))); % fitting error should only be based on the fitting of PDF/CDF/Moments and not on the penalty for the number of states of an Erlang branch
fprintf('lsqlin finished: norm2_error is %E   exitflag is %d \n',norm2_error,exitflag);
printtime();


epsilonpivec = 1.0e-4;
pivec_invalid = ( (sum(abs(pivecres)) > (1+epsilonpivec))  || (sum(abs(pivecres)) < (1-epsilonpivec))  || (min(pivecres) < -epsilonpivec));

fprintf('fitpdf_and_cdf_withmoments:  pivec_invalid = %d, pivecsum = %E, min(pivecres) = %E\n',pivec_invalid,sum(abs(pivecres)),min(pivecres));

if ((exitflag < 0) || (pivec_invalid > 0))
  norm2_error = 1.0e+90;
  exitflag = -2;
end


delepsilon = 1.0e-12;  %   1.0e-15;	 
pivecres(pivecres < delepsilon) = 0;  % neglect nearly zero entries for making plots
                                   % we don't use delete_pi_zero_entries here in order
                                   % to keep the dimension for multiplication with Ppdf/Pcdf/Mbestall matrices
                                   % in call for plot_of_result.

if ((norm2_error <= norm2_error_threshold) && (exitflag >= 0) && (pivec_invalid == 0))  % plot result only if it is valid and better than threshold
    % plot_of_result(tracepdf', (pivecres'*Ppdf), tracecdf', (pivecres'*Pcdf), tracemoments, (pivecres'*Mbestall'), pdftvalues', find_index(pdftvalues',2.0), cdftvalues', find_index(cdftvalues',2.0),jvm);
    plot_of_result(tracepdf', (pivecres'*Ppdf), tracecdf', (pivecres'*Pcdf), tracemoments, (pivecres'*Mbestall'), pdftvalues', length(pdftvalues), cdftvalues', length(cdftvalues),jvm);
end

pivecres = pivecres';
[M,pivec,Ak,inivec,N,C,~] = delete_pi_zero_entries_fast(Mbestall,pivecres,Akbestall,inivecbestall,no_list_elems,C,delepsilon);
delepsilon2 = 0.001/N;
% [M,pivec,Ak,inivec,N,C,~] = delete_less_relevant_components(M,pivec,Ak,inivec,N,tracemoments,C,delepsilon2);
tmpvec = C*pivec'-d;
norm2_error = norm(tmpvec(1:(length(tmpvec)-1))); % fitting error should only be based on the fitting of PDF/CDF/Moments and not on the penalty for the number of states of an Erlang branch
fprintf('lsqlin finished: norm2_error after deletion is %E   exitflag is %d \n',norm2_error,exitflag);
printtime();


