% File <fittracedist.m>
%
% Author: Falko Bause
% Date: June 2017   changed February 2018 (storage efficient representation of Ak)
%
% This file is part of the MATLAB functions for fitting mixture distributions
% Copyright (C) Falko Bause, Department of Computer Science, TU Dortmund, 2016.
%
% MAIN FUNCTION!!!
%   This function fits a trace to a mixture of Erlang distributions (Hyper-Erlang distribution).
%     If the code is slightly changed additionally deterministic distributions are considered in the mixture.
%
% INPUT:
%     tracefile                              (absolute) file name of trace file. Format is a list of values (one value per line), e.g.:
%                                                           406.338
%                                                           406.341
%                                                           406.348
%
%     nvars1, nvars2                         determine approximate number of variables for NNLS problem solved in fitpdf_and_cdf_withmoments.m
%                                            nvars1 is used for the Farey list for parameter lambda, nvars2 for the different number of states (klist), i.e. number of phases of the Erlang branches
%
%     no_cdfpoints                           number of points at which the CDF values for trace and Erlang/deterministic distributions are used for fitting
%     Kdist                                  number of moments used for fitting
%     quantilesno                            number of quantiles used for determining intervals in which Farey points are constructed;
%                                            essentially helps to construct better suited sets of distributions
%     threshold_for_pre_analysis_of_trace    if > 0  a preanalysis of the trace is performed, all values of the trace whose (relative) number of occurences
%                                            is above the given threshold value are used as mean values for the deterministic/Erlang distributions
%     threshold_norm2_error                  the fitting is aborted once the norm-2 error of NNLS fitting is below the given value
%     novariates_result                      if > 0 a trace is generated using the resultant Hyper-Erlang/dtm. distribution, the number of
%                                            generated random variates is at least novariates_result. The trace is stored in file 'shotgun.trace'.
%                                            NOTE that the generated trace is to some degree sorted for a quicker generation, so that it can
%                                            only be used for comparisons concerning distribution figures !!!! (e.g. autocorrelation will nearly
%                                            always be non zero)
%     pdfcdffit_only                         if > 0 only PDF/CDF is fitted, moment fitting is usually very bad in this case
%                                            if == 0 PDF/CDF   AND  moments are simultaneously fitted
%     jvm                                    if == 1 this should indicate that matlab has been called with active Java virtual machine (i.e. no option -nojvm is used).
%                                              In this case additional PDF files are created showing (interim, final resp.) fitting results.
%                                            if == 0 it is assumed that matlab has been called with option -nojvm and thus no figures are created, since for creation Java VM is needed.
%                                              Usually option -nojvm  reduces fitting times significantly.
%     nostatepenalty                         is a penalty factor disadvantaging Erlang branches with many states
%
%
% OUTPUT:
%     pivec                                  vector giving the mixture probabilities of selecting a deterministic/Erlang distribution specified in vectors inivec,Ak
%     M                                      is the moment matrix (should be consistent with pivec, Ak, inivec)
%     Ak                                     list of deterministic/Erlang distributions
%     inivec                                 starting probabilities of distributions in Ak. Since we use deterministic (here inivec(i) is irrelevant) and
%                                            Erlang distributions, inivec should always be a vector similar to (1,0,...)
%     deterministic_free                     is 1 if the list of Ak distributions does NOT contain a deterministic distribution, 0 otherwise
%     norm2_error                            is the best resultant Norm-2 error of (possibly several) NNLS problems solved by function lsqlin


function [pivec,M,Ak,inivec,Nlag1states,deterministic_free,norm2_error] = fittracedist(tracefile, nvars1, nvars2, no_cdfpoints, Kdist, quantilesno, threshold_for_pre_analysis_of_trace,threshold_norm2_error,novariates_result,pdfcdffit_only,jvm,nostatepenalty)

% DEFAULTS (in case parameter is not specified in function call)

if ~exist('tracefile', 'var')
    fprintf('At least a file name must be specified for the call of fittracedist!!!  For other parameters default values are specified.  \n\n');
    fprintf('Function prototype is \n [pivec,M,Ak,inivec,Nlag1states,deterministic_free,norm2_error] = fittracedist(tracefile, nvars1, nvars2, no_cdfpoints, Kdist, quantilesno, threshold_for_pre_analysis_of_trace,threshold_norm2_error,novariates_result,pdfcdffit_only,jvm,nostatepenalty)\n\n\n');
    
    fprintf('Possible INPUT parameters of fittracedist call are:\n');
    fprintf('tracefile                             (absolute) file name of trace file. Format is a list of values (one value per line), e.g.: \n   406.338\n   406.341\n   406.348\n\n');
    fprintf('nvars1, nvars2                        determine approximate number of variables for NNLS problem.\n');
    fprintf('                                         nvars1 is used for parameter lambda, nvars2 for the number of phases of the Erlang branches.\n');
    fprintf('no_cdfpoints                          number of points at which the CDF values for trace and Erlang distributions are used for fitting.\n');
    fprintf('Kdist                                 number of moments used for fitting\n');
    fprintf('quantilesno                           number of quantiles used for determining intervals in which Farey points are constructed;\n');
    fprintf('                                         essentially helps to construct better suited sets of distributions.\n');
    fprintf('threshold_for_pre_analysis_of_trace   if > 0  a preanalysis of the trace is performed, all values of the trace whose (relative) number of occurences\n');
    fprintf('                                         is above the given threshold value are used as mean values for the Erlang distributions.\n');
    fprintf('threshold_norm2_error                 the fitting is aborted once the norm-2 error of NNLS fitting is below the given value.\n');
    fprintf('                                         is only relevant if the NNLS solver fails for the initial overall problem definition\n');
    fprintf('novariates_result                     if > 0 a trace is generated using the resultant Hyper-Erlang distribution, the number of\n');
    fprintf('                                         generated random variates is at least novariates_result. The trace is stored in file shotgun.trace.\n');
    fprintf('                                         NOTE that the generated trace is to some degree sorted for a quicker generation, so that it can\n');
    fprintf('                                         only be used for comparisons concerning distribution figures !!!! \n');
    fprintf('                                          (e.g. autocorrelation will nearly always be non zero)\n');
    fprintf('pdfcdffit_only                        if > 0 only PDF/CDF is fitted, moment fitting is usually very bad in this case\n');
    fprintf('                                      if == 0 PDF/CDF   AND  moments are simultaneously fitted\n');
    fprintf('jvm                                   if == 1 this should indicate that matlab has been called with active Java virtual machine (i.e. no option -nojvm is used).\n');
    fprintf('                                         In this case additional PDF files are created showing (interim, final resp.) fitting results.\n');
    fprintf('                                      if == 0 it is assumed that matlab has been called with option -nojvm and thus no figures are created, since for creation Java VM is needed.\n');
    fprintf('                                         Usually option -nojvm reduces fitting times significantly.\n');
    fprintf('nostatepenalty                        is a penalty factor disadvantaging Erlang branches with many states (recommended is some integer value from interval [0..20])\n\n');

    fprintf('OUTPUT parameters are:\n');
    fprintf('pivec                                 vector giving the mixture probabilities of selecting an Erlang distribution specified in vectors inivec,Ak\n');
    fprintf('M                                     is the moment matrix\n');
    fprintf('Ak                                    list of Erlang distributions\n');
    fprintf('inivec                                starting probabilities of distributions in Ak. Since we use Erlang distributions, inivec_i is here a vector (1.0 , number_of_states)\n');
    fprintf('deterministic_free                    is 1 if the list of Ak distributions does NOT contain a deterministic distribution, 0 otherwise.\n');
    fprintf('                                         should always be the case in the current version of function fittracedist.\n');
    fprintf('norm2_error                           is the (best) resultant Norm-2 error of (possibly several) NNLS problem(s) solved by function lsqlin.\n\n');
    
    return;
end
if ~exist('nvars1', 'var')
    nvars1 = 30;
end
if ~exist('nvars2', 'var')
    nvars2 = 20;
end
if ~exist('no_cdfpoints', 'var')
    no_cdfpoints = 1000;
end
if ~exist('Kdist', 'var')
    Kdist = 10;
end
if ~exist('quantilesno', 'var')
    quantilesno = 10;
end
if ~exist('threshold_for_pre_analysis_of_trace', 'var')
    threshold_for_pre_analysis_of_trace = 0.0;
end
if ~exist('threshold_norm2_error', 'var')
    threshold_norm2_error = 0.00001;
end
if ~exist('novariates_result', 'var')
    novariates_result = 0;
end
if ~exist('pdfcdffit_only', 'var')
    pdfcdffit_only = 0;
end
if ~exist('jvm', 'var')
    jvm = 1;
end
if ~exist('nostatepenalty', 'var')
    nostatepenalty = 0.0;
end

assumeandprint(exist(tracefile, 'file') == 2,'trace file does not exist\n'); % tracefile should exist
assumeandprint(nvars1 >= 10,'nvars1 should be >= 10\n'); % some minimum value. would fitting make sense with less ...??
assumeandprint(nvars2 >= 10,'nvars2 should be >= 10\n'); % some minimum value. would fitting make sense with less ...??
assumeandprint(no_cdfpoints >= 50,'no_cdfpoints should be >= 50\n');
% as a rule of thumb no_cdfpoints > nvars1*nvars2 is a good choice, which should result in adequate NNLS problems where the number of equations exceeds the number of variables.
% Note that also the points at which PDF values are used in the NNLS problem specification will also contribute to the number of equations. This number can not be predicted without analysis of the trace,
% since the Freedman/Diaconis approach followed by a possible correction is used, see file epdf_and_cdfatvals.m.
assumeandprint(Kdist > 1,'Kdist should be > 1\n');
assumeandprint(quantilesno > 2,'quantilesno should be > 2\n');
assumeandprint(threshold_for_pre_analysis_of_trace >= 0,'threshold_for_pre_analysis_of_trace is negative!\n');
assumeandprint(threshold_norm2_error >= 0,'threshold_norm2_error is negative!\n');
assumeandprint(novariates_result >= 0,'novariates_result is negative!\n');

% DEFINITIONS
%      Note that nearly all vectors used are row vectors and are transposed where needed

% options = optimset('Display','notify','Diagnostics','off','MaxIter',100000,'TolFun',1.0e-20,'TolX',1.0e-20);
% options = optimset('Display','notify','Diagnostics','off','MaxIter',10000,'TolFun',1.0e-20,'TolX',1.0e-20);
options = optimset('Display','notify','Diagnostics','off','MaxIter',1000,'TolFun',1.0e-20,'TolX',1.0e-20);




epsilon =  1.0e-6;

% N Number of components
% Kdist = number of moments

% K = Kdist;
% The 0th moment is always one, so later in matrix M and for moment vectors
% we use K+1 moments.


Nlag1states = 0;  % later used for the total number of states in the mixture distribution

Ak = {};        % generators of the components
inivec = {};    % initial probability vectors of the components
N = 0;

% Initialize matrices M  containing the moments of the components
M = [];
% Initialize vector pivec  later used for the mixture probabilities
pivec = (1);
pivec_initialguess = []; % for later NNLS calls in function fitpdf_and_cdf_withmoments.m

global zeit;
zeit = tic;
fprintf('Function fittracedist started!    ');
printtime();


% Read data of tracefile
fid = myfopen(tracefile);
tr = textscan(fid,'%f');
trace = tr{1};
fclose(fid);

if (min(trace) < 0)
   fprintf('Some trace elements are NEGATIVE!. Since Hyper-Erlang Distributions are defined only on the positive domain, negative elements will be deleted!!!\n') 
   Ltr = length(trace);
   trace = trace(trace > 0);
   fprintf('Original trace length = %d, length after deleting non-positive elements = %d \n',Ltr,length(trace));
end

% STEPS OF ALGORITHM:

[~,fname,~] = fileparts(tracefile);
printtrace_characteristics( tracefile, trace, fname, quantilesno );

distmoments = tracedist_start_from_zero(trace, Kdist+1);

maxdistelem = max(trace);
mindistelem = min(trace);
meandistelem = mean(trace);
mediandistelem = median(trace);
% fprintf('mindistelem = %E, maxdistelem = %E, meandistelem = %E, mediandistelem = %E\n',mindistelem,maxdistelem,meandistelem,mediandistelem);

% maximum value of a components  moment vector entry
maxval =  realmax/1.0e+10;  % or 1.0e+90;

if (max(distmoments) > maxval)
    fprintf('Moments of trace are greater than maxval (Max. Moments = %E; maxval = %E). Reduce parameter Kdist, i.e. the number of moments to fit! (Kdist = %d)\n',max(distmoments),maxval,Kdist);
    exit;
end
if (max(trace) > maxval)
    fprintf('Maximum of trace is greater than maxval (Max. Trace = %E; maxval = %E). Delete such large values/outliers from trace before fitting, otherwise this fitting tool can not be used.\n',maxdistelem,maxval);
    exit;
end

norm2_error = 1.0e+90;
norm2_errortmp = 1.0e+90;

pivectmp = pivec;
Mtmp = M;
inivectmp = inivec;
Aktmp = Ak;
Ntmp = N;

fprintf('quantilesno = %d\n',quantilesno);
qv = quantile(trace,quantilesno-2);

fprintf('Possibly adding extra values to the quantile vector, if maximum/minimum of trace differ orders of magnitude from quantiles of trace.\n');
fprintf('This might e.g. be the case if the trace contains outliers or kurtosis/skewness is high. \n');
qvprint=sprintf('%E ', qv);
fprintf('Current qv is \n      %s\n',qvprint);
fprintf('Note: It might be necessary to increase the number of cdfpoints to get adequate NNLS problem definitions.\n\n');

qv = addquantilevalues(qv,trace,quantilesno);
qv = unique(horzcat(mindistelem,qv,maxdistelem));
qv = horzcat(qv,meandistelem,mediandistelem);
qvepsilon = (maxdistelem-mindistelem)/length(trace);
qv = unique_values(qv,qvepsilon);
qvprint=sprintf('%E ', qv);
fprintf('Final qv used is \n      %s\n\n',qvprint);

cdftvalues = determine_cdftvalues(trace,no_cdfpoints);
[distpdf, distcdf, pdftvalues, cdftvalues] = epdf_and_cdfatvals(trace,cdftvalues);
% [distpdf, distcdf, pdftvalues, cdftvalues] = epdf_and_cdfatvals_with_ksdensity(trace,cdftvalues);

% determine values vector for nvars1:
no_farey_order = determine_order_of_number_of_farey_points(round(nvars1/(length(qv)-1)));
farvalues = farey_points(no_farey_order)

values = [];
for i=1:(length(qv)-1)
    values = horzcat(values,((qv(i+1)-qv(i))*farvalues+qv(i)));
end
if (threshold_for_pre_analysis_of_trace > 0)
    % Do a preanalysis of the trace and insert the most occurring values into array values.
    fprintf('Performing preanalysis of trace and possibly inserting extra values!\n');
    fprintf('Threshold is %E\n',threshold_for_pre_analysis_of_trace);
    valuestmp = find_values_above_threshold(trace,threshold_for_pre_analysis_of_trace);
    fprintf('The following values are additionally inserted into the values vector: ');
    setdiff(valuestmp,values)
    values = horzcat(values,valuestmp);
end

values = unique_values(values,epsilon); % step is necessary, since values from preanalysis of trace (and former steps for
					% determining qv vector) might approx. coincide with values generated from farey values
                                        % resulting in badly conditioned NNLS problems

fprintf('values vector: \n'); matrixoutputexpo(values); fprintf('\n');
fprintf('length of values vector is: %d   minimum difference is %E   maximum difference is %E    quotient is %E\n',length(values),min(diffvec(values)),max(diffvec(values)),max(diffvec(values))/max(10^(-12),abs(min(diffvec(values)))));
fprintf('quotient/median(values): %E\n\n\n',(max(diffvec(values))/max(10^(-12),abs(min(diffvec(values))))/median(values)));

% determine values for nvars2:
% for a large number of states an Erlang dist approaches a normal dist.
% The maximum value of the pdf of a normal dist is at the mean with
% function value (sqrt(2*pi)*sigma)^(-1) with sigma = sqrt(k)/lambda.
% Thus taken a value of pdftvalues as a mean the corresponding function
% value distpdf can only be reached if the number of states/phases is
% greater than kmax
kmax = max(1,round(2*pi*(max(distpdf.*pdftvalues'))^2));
fprintf('kmax = %d\n',kmax);

% Determine scaling factor betaval such that kmax and also some larger values are
% in the list of kvalues which determines the number of states of the Erlang distributions in Mfield
betaval = 1/100; % only some initial value used as scaling factor for kvalues
maxklist = [1];
while ((min(100*kmax,10^8) >= max(maxklist))  && (max(maxklist) < 10^8))  % 10^8 for technical reasons (limiting maximum matrix size)
    betaval = betaval * 2.0;
    ORDERMAX = round(determine_order_for_kvalue_list(max(1,30),betaval)); % 30 is just an example value for nvars2
    % essentially here a fixed value is used to ensure that betaval only depends on trace data (kmax) and
    % the approximate maximum list entry of 10^8
    % this implies that maxklist(nvars2) \subseteq maxklist(nvars2+1) and the fitting result should improve with
    % increasing nvars2 values.  Note: The argumentation here is similar to the one for nvars1 using farey points
    % and a transformation based on sample quantiles.
    maxklist = kvalues(ORDERMAX,betaval); % is the maximum klist
end  % whole while loop is a heuristics to find an appropriate value for transformation parameter betaval

fprintf('betaval = %E\n',betaval);
ORDERMAX = round(determine_order_for_kvalue_list(max(1,nvars2),betaval))
maxklist = kvalues(ORDERMAX,betaval) % is the maximum klist

order = ORDERMAX;
oldklistlength = 0;
exitflag = -2;
fitsolution_found = 0;


% In the following we start solving the NNLS problem with the maximum klist (maxklist).
% If this fails, we switch the approach and solve several NNLS problems always using the same
% values vector from above and increasing the length of klist. Essentially
% only the last iteration should be relevant (with the maximum klist, as we tried in the first step) since
% then the set of Erlang distributions (Mfield) comprises all Erlangs sets
% from former iterations. Unfortunately due to bad performance of lsqlin or
% due to a badly conditioned NNLS problem it might happen that no solution
% is found for the maximum klist. Therefore we keep the best solution from former iterations,
% since usually at least one of the former problems gives a reasonable
% solution.

first_run = 1;

while ( (fitsolution_found <= 0)  && (norm2_error > threshold_norm2_error) && (order <= ORDERMAX) )
    
    klist = kvalues(order,betaval)
    newklistlength = length(klist);
    
    if (newklistlength > oldklistlength)  % since it happens that increasing order by only 1 is not sufficient to increase klist
        % and thus the same NNLS problem as before would be solved; essentially relevant if the first lsqlin run fails.
        oldklistlength = newklistlength;
        
        [Mfull,Akfull,inivecfull] = determine_Mfield_erlang_bounds(distmoments,klist, Kdist, values, maxval);
        Mcounter = size(Mfull,2);
        
        % Use pivec of current best solution of former NNLS call in function fitpdf_and_cdf_withmoments as initial guess for new NNLS problem if possible:
        if ((length(pivec) <= 1) || (Mcounter <= length(pivec)))
	    pivec_initialguess = (ones(Mcounter,1) / Mcounter); 
        else
            pivec_initialguess = vertcat(pivec',(zeros(Mcounter-length(pivec),1)));
        end
        
        [pivectmp,Mtmp,inivectmp,Aktmp,Ntmp,norm2_errortmp,exitflag] = fitpdf_and_cdf_withmoments(distmoments, distpdf', distcdf', pdftvalues, cdftvalues, pivec_initialguess, inivecfull, Akfull, Mfull, Kdist, norm2_error, options,pdfcdffit_only,jvm,nostatepenalty);
        fprintf('***** klist order = %d, Ntmp = %d, no_states = %d, norm2_errortmp = %E  norm2_error = %E \n',order,Ntmp,count_states(inivectmp,Ntmp),norm2_errortmp, norm2_error);
        
        % if ((norm2_errortmp < norm2_error) && ( ((exitflag >= 0) && (first_run <= 0)) || ((first_run >= 1) && (norm2_errortmp < 1.0e-2) && (norm2_errortmp < norm2_error) && (exitflag >= 1)) ) )
        if ((norm2_errortmp < norm2_error) && ( ((exitflag >= 0) && (first_run <= 0)) || ((first_run >= 1) && (norm2_errortmp < norm2_error) && (exitflag >= 1)) ) )
            if (first_run >= 1)
                fitsolution_found = 1;
            end
            norm2_error = norm2_errortmp;
            pivec = pivectmp;
            M = Mtmp;
            inivec = inivectmp;
            Ak = Aktmp;
            N = Ntmp;
            fprintf('Current number of components:  N = %d\n\n',N);
        else
            if (first_run >= 1)
                first_run = 0;
                order = 0;
                oldklistlength = 0;
                if ((exitflag >= 0) && (norm2_errortmp < norm2_error))
                    norm2_error = norm2_errortmp;
                    pivec = pivectmp;
                    M = Mtmp;
                    inivec = inivectmp;
                    Ak = Aktmp;
                    N = Ntmp;
                    fprintf('Current number of components:  N = %d\n\n',N);
                end
            else
                fprintf('No improvement of norm2_error.\n\n');
            end
        end
    end
    order = order + 1;
end

order = order - 1;

[M,pivec,Ak,inivec,N,norm2_error,~,exitflag] = reduce_number_of_components(M,pivec,Ak,inivec,N,distmoments,distpdf,distcdf,pdftvalues,cdftvalues,Kdist,norm2_error,options,pdfcdffit_only,jvm,nostatepenalty,exitflag);


fprintf('Function fittracedist: End of fitting!!!    ');
fittingtime=printtime();

if (norm2_error > 1.0e+70)
    fprintf('fittracedist: lsqlin did NOT find an adequate solution!! fittracedist is now returning with NO result!!!  last exitflag = %d\n',exitflag);
    pivec = [];
    M = [];
    Ak = [];
    inivec = [];
    Nlag1states = -1;
    deterministic_free = -1;
    commandstrdel = ['\rm -f shotgun*.pdf shotgun.trace shotgun.profido'];
    [status,result] =  system(commandstrdel);
    return;
end


Nlag1states = 0;
for i=1:N
    if (inivec{i}(1,2) ~= 123456789)
      Nlag1states = Nlag1states + inivec{i}(1,2);
    else
      Nlag1states = Nlag1states + 1;  % deterministic dist is counted as a single distribution
    end
end

epsilonpivec = 1.0e-4;

if ( (sum(abs(pivec)) > (1+epsilonpivec))  || (min(pivec) < -epsilonpivec))
    fprintf('WRONG SOLUTION:  pivec vector is invalid!!!\n\n pivec = ');
    matrixoutputexpo(pivec);
    fprintf('(sum(abs(pivec)) = %E\n',(sum(abs(pivec))));
    fprintf('\n\n aborting fittracedist ... \n\n');
    Nlag1states = -1;
    deterministic_free = -1;
    return;
end

if (norm2_error > 1.0e+70)
    fprintf('fittrace:  NO solution found!!  norm2_error %E\n',norm2_error);
    fprintf('\n\n aborting fittracedist ... \n\n');
    Nlag1states = -1;
    deterministic_free = -1;
    return;
end


% Transfomation into a MAP representation for checking distribution with
% ProFiDo. Does only work if there is no deterministic distribution
% involved:

deterministic_free = check_deterministic_free(Ak,N);
if ((deterministic_free > 0) && (Nlag1states < 1000))   % using ProFiDo's plot functionality is only recommended for smaller state spaces, we use 1000 here.
    writeprofido_hypererlang(pivec,Ak);
end

if (N > 0)
    pivec_threshold = 0.0001;
    
    resultantrelerror = max(abs((pivec*M'-distmoments)./distmoments));  % relative Maximum-Norm
    fprintf('\n\n *********************************************************************************************************\n');
    fprintf('N=%E, resultantrelerror (Max-Norm) = %E\n',N,resultantrelerror);
    fprintf('distmoments = \n');
    matrixoutputexpo(distmoments);
    fprintf('moments of fitted distribution = \n');
    matrixoutputexpo(pivec*M');
    fprintf('pivec*M  (only for pivec[i] >= %E) = \n',pivec_threshold);
    matrixoutputexpo(set_to_zero(pivec,pivec_threshold)*M');
    fprintf('pivec*M  (only for pivec[i] >= %E) = \n',(10.0*pivec_threshold));
    matrixoutputexpo(set_to_zero(pivec,(10.0*pivec_threshold))*M');
    fprintf('relative errors of moments = \n');
    matrixoutputexpo(abs((pivec*M'-distmoments)./distmoments));
    relerrormoments=abs((pivec*M'-distmoments)./distmoments);
    fprintf('ratio = \n');
    ratiovec = abs((pivec*M')./distmoments);
    maxratio = max(ratiovec); minratio = min(ratiovec);
    maxratiodiff = maxratio-minratio;
    matrixoutputexpo(ratiovec);
    fprintf('==> maxratiodiff = %E\n',maxratiodiff);
    fprintf('resultant pivec:\n');
    matrixoutputexpo(pivec);
    fprintf('\n\n\nRESULT OF FITTING IS:\n');
    output_deterministic_Aks(Ak, N, pivec);
    output_erlang_Aks(Ak, N, pivec);
    fprintf('\n');
end

printtime();

if ((jvm > 0) | (novariates_result > 0))
  % Calculating log likelihood might take some time, so we skip this step if matlab has been called with option -nojvm  or no variates should be generated just to reduce program execution time
   loglikval = loglikelihood_dtm(pivec,Ak,trace');
   % fprintf('\n\nLOG Likelihood value is     %E\n\n\n',loglikval);
   fprintf('\n\nLOG Likelihood =  %E, trace length = %d, loglike/trace_length = %E, Nlag1states = %d\n\n\n',loglikval,length(trace),loglikval/length(trace),Nlag1states);
else
   fprintf('\n\nNlag1states = %d\n\n\n',Nlag1states);
end

prettyprint(tracefile,fittingtime,N,pivec,Ak,relerrormoments,norm2_error);

fprintf('\n\nSHORT SUMMARY File: %s NoComp: %d States: %d Norm-2-Error: %E MaxRelErrorMoments: %E Fitting time (sec.): %f\n\n',fname,N,Nlag1states,norm2_error,max(relerrormoments),fittingtime);

if ((deterministic_free > 0) && (novariates_result > 0))
    % novariates_result should be large enough, so that also the least probable Erlang branch will contribute significantly often to the resultant trace
    % number is less or equal to 1.0e+8 to prevent memory overflow in ProFiDo's plot program
    
    novariates_result_recommended = round(1000*(1.0/min(pivec)));
    fprintf('Number of variates recommended = %d, number of variates requested = %d\n',novariates_result_recommended,novariates_result);
    novariates_result = round(min(1.0e+8,max(novariates_result_recommended,novariates_result)))
    generate_sorted_hypererlang_random_variates(pivec,inivec,Ak,novariates_result,'shotgun.trace');
    
end

fprintf('\n\n eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeend        ');
printtime();



