% File <reduce_number_of_components.m>
%
% Author: Falko Bause
% Date: August 2019
%
% This file is part of the MATLAB functions for fitting mixture distributions
% Copyright (C) Falko Bause, Department of Computer Science, TU Dortmund, 2016.
%


function [M,pivec,Ak,inivec,N,norm2_error,component_deleted,exitflag] = reduce_number_of_components(M,pivec,Ak,inivec,N,distmoments,distpdf,distcdf,pdftvalues,cdftvalues,Kdist,norm2_error,options,pdfcdffit_only,jvm,nostatepenalty,exitflag)
% trying to reduce the number of components without loosing fitting
% quality. If the moments are not changed significantly when eliminating
% less relevant components (identified by pivec entry being less than a
% theshold), we do a final fitting step by eliminating those components
% with lower pivec values and fitting the pivec vector only for the
% remaining components. If this fails the input parameters are
% returned, i.e. we return the former fitting result.

fprintf('Function reduce_number_of_components start.\n');

xpm = 200; % 50;  % number of permilles moment fitting may be worsened by removing components
nx = 10;   % number of percent overall fitting may be worsened by removing components
normfactor = (1 + nx * 0.01);
component_deleted = 0;

if ((N > 0) && (norm2_error < 1))   % start reduction only if former fitting steps were successful
    
    maxdistmerr = 1.0;
    pivec_threshold = 0.1;
    fitmoments = pivec*M';
    
    while (maxdistmerr > (xpm/1000.0)) % removing corresponding components would change the fitted moment vector elements by at most xpm permille
        moments = set_to_zero(pivec,pivec_threshold)*M';
        maxdistmerr = max(abs((moments-fitmoments)./fitmoments));
        pivec_threshold = pivec_threshold / 10.0 ;
    end
    
[Mtmp,pivectmp,Aktmp,inivectmp,~,~,component_deleted]= delete_pi_zero_entries_fast(M,pivec,Ak,inivec,N,eye(N),(10.0*pivec_threshold));
    
    if (size(Mtmp,2) < N)   % fitting attempt only necessary if components can be deleted
        [pivectmp,Mtmp,inivectmp,Aktmp,Ntmp,norm2_errortmp,exitflagtmp] = fitpdf_and_cdf_withmoments(distmoments, distpdf', distcdf', pdftvalues, cdftvalues, pivectmp', inivectmp,Aktmp,Mtmp,Kdist, (norm2_error*normfactor), options,pdfcdffit_only,jvm,nostatepenalty);
        
        if ((exitflagtmp >= 0) && (norm2_errortmp < (norm2_error * normfactor)))
            % fitting was successful;
            % first output the former result before removing less relevant components
            % provided we have really reduced the number of components
            if (N > Ntmp)
                pivec_thresholdtmp = 0.0001;
                
                resultantrelerror = max(abs((pivec*M'-distmoments)./distmoments));  % relative Maximum-Norm
                fprintf('\n\n INTERIM RESULT BEGIN (this is the fitting result before reducing number of components)\n');
                fprintf('N=%E, resultantrelerror (Max-Norm) = %E\n',N,resultantrelerror);
                fprintf('distmoments = \n');
                matrixoutputexpo(distmoments);
                fprintf('moments of fitted distribution = \n');
                matrixoutputexpo(pivec*M');
                fprintf('pivec*M  (only for pivec[i] >= %E) = \n',pivec_thresholdtmp);
                matrixoutputexpo(set_to_zero(pivec,pivec_thresholdtmp)*M');
                fprintf('pivec*M  (only for pivec[i] >= %E) = \n',(10.0*pivec_thresholdtmp));
                matrixoutputexpo(set_to_zero(pivec,(10.0*pivec_thresholdtmp))*M');
                fprintf('relative errors of moments = \n');
                matrixoutputexpo(abs((pivec*M'-distmoments)./distmoments));
                relerrormoments=abs((pivec*M'-distmoments)./distmoments);
                fprintf('ratio = \n');
                ratiovec = abs((pivec*M')./distmoments);
                maxratio = max(ratiovec); minratio = min(ratiovec);
                maxratiodiff = maxratio-minratio;
                matrixoutputexpo(ratiovec);
                fprintf('==> maxratiodiff = %E\n',maxratiodiff);
                fprintf('resultant pivec:\n');
                matrixoutputexpo(pivec);
                fprintf('\n\n\nResult of fitting before trying to reduce the number of components:\n');
                output_deterministic_Aks(Ak, N, pivec);
                output_erlang_Aks(Ak, N, pivec);
                fprintf('\n INTERIM RESULT:  norm2_error before reduction: %E   after reduction  %E\n',norm2_error,norm2_errortmp);
                fprintf('\n INTERIM RESULT END (number of components: before reduction %d  after reduction (see below) %d  deleted %d)\n',N,Ntmp,N-Ntmp);
                
                norm2_error = norm2_errortmp;
                M = Mtmp;
                pivec = pivectmp;
                Ak = Aktmp;
                inivec = inivectmp;
                N = Ntmp;
                exitflag = exitflagtmp;
            end
        end
    end
end
fprintf('Function reduce_number_of_components end.\n');
